import pytest
import torchvision
import torch
import torch.nn as nn
import torch.nn.functional as F


class SimpleNet(nn.Module):
    def __init__(self):
        super(SimpleNet, self).__init__()
        self.conv1 = nn.Conv2d(1, 6, 3)
        self.conv2 = nn.Conv2d(6, 16, 3)
        self.fc1 = nn.Linear(16 * 6 * 6, 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, 10)

    def forward(self, x):
        x = F.max_pool2d(F.relu(self.conv1(x)), (2, 2))
        x = F.max_pool2d(F.relu(self.conv2(x)), 2)
        x = x.view(-1, self.num_flat_features(x))
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x

    def num_flat_features(self, x):
        size = x.size()[1:]
        num_features = 1
        for s in size:
            num_features *= s
        return num_features


class MultiInputNet(nn.Module):
    def __init__(self):
        super(MultiInputNet, self).__init__()
        self.features = nn.Sequential(
            nn.Conv2d(1, 1, kernel_size=3, stride=1, padding=1), nn.ReLU()
        )

    def forward(self, x, y):
        x1 = self.features(x)
        x2 = self.features(y)
        return x1, x2


class SimpleBatchNormModel(nn.Module):
    def __init__(self):
        super(SimpleBatchNormModel, self).__init__()
        hidden = 15
        input_size = 20
        self.main = nn.Sequential(
            nn.Linear(input_size, hidden, bias=False),
            nn.BatchNorm1d(hidden),
            nn.Linear(hidden, hidden, bias=False),
            nn.BatchNorm1d(hidden),
            nn.Linear(hidden, 1),
        )

    def forward(self, input):
        output = self.main(input)
        return output


class Generator(nn.Module):
    def __init__(self, ngpu=1):
        super(Generator, self).__init__()
        nz = 100
        ngf = 64
        nc = 3
        self.ngpu = ngpu
        self.main = nn.Sequential(
            nn.ConvTranspose2d(nz, ngf * 8, 4, 1, 0, bias=False),
            nn.BatchNorm2d(ngf * 8),
            nn.ReLU(True),
            nn.ConvTranspose2d(ngf * 8, ngf * 4, 4, 2, 1, bias=False),
            nn.BatchNorm2d(ngf * 4),
            nn.ReLU(True),
            nn.ConvTranspose2d(ngf * 4, ngf * 2, 4, 2, 1, bias=False),
            nn.BatchNorm2d(ngf * 2),
            nn.ReLU(True),
            nn.ConvTranspose2d(ngf * 2, ngf, 4, 2, 1, bias=False),
            nn.BatchNorm2d(ngf),
            nn.ReLU(True),
            nn.ConvTranspose2d(ngf, nc, 4, 2, 1, bias=False),
            nn.Tanh(),
        )

    def forward(self, input):
        if input.is_cuda and self.ngpu > 1:
            output = nn.parallel.data_parallel(
                self.main, input, range(self.ngpu)
            )
        else:
            output = self.main(input)
        return output


class LSTMTagger(nn.Module):
    def __init__(self, embedding_dim, hidden_dim, vocab_size, tagset_size):
        super(LSTMTagger, self).__init__()
        self.hidden_dim = hidden_dim
        self.word_embeddings = nn.Embedding(vocab_size, embedding_dim)
        self.lstm = nn.LSTM(embedding_dim, hidden_dim)
        self.hidden2tag = nn.Linear(hidden_dim, tagset_size)

    def forward(self, sentence):
        embeds = self.word_embeddings(sentence)
        lstm_out, _ = self.lstm(embeds.view(len(sentence), 1, -1))
        tag_space = self.hidden2tag(lstm_out.view(len(sentence), -1))
        tag_scores = F.log_softmax(tag_space, dim=1)
        return tag_scores


class SimpleConv(nn.Module):
    def __init__(self):
        super(SimpleConv, self).__init__()
        self.features = nn.Sequential(
            nn.Conv2d(1, 1, kernel_size=3, stride=1, padding=1), nn.ReLU()
        )

    def forward(self, x, y):
        x1 = self.features(x)
        x2 = self.features(y)
        return x1, x2


class Autoencoder(nn.Module):
    def __init__(self):
        super(Autoencoder, self).__init__()
        self.encoder = nn.Sequential(
            nn.Conv2d(3, 6, kernel_size=5),
            nn.ReLU(True),
            nn.Conv2d(6, 16, kernel_size=5),
            nn.ReLU(True),
        )
        self.decoder = nn.Sequential(
            nn.ConvTranspose2d(16, 6, kernel_size=5),
            nn.ReLU(True),
            nn.ConvTranspose2d(6, 3, kernel_size=5),
            nn.ReLU(True),
            nn.Sigmoid(),
        )

    def forward(self, x):
        x = self.encoder(x)
        x = self.decoder(x)
        return x


@pytest.fixture(scope='session')
def simple_model():
    net = SimpleNet()
    return net


@pytest.fixture(scope='session')
def mobilenet():
    model = torchvision.models.mobilenet_v2()
    return model


@pytest.fixture(scope='session')
def multi_input_net():
    model = MultiInputNet()
    return model


@pytest.fixture(scope='session')
def lstm_tagger():
    model = LSTMTagger(6, 6, 5, 3)
    return model


@pytest.fixture(scope='session')
def netgenerator():
    model = Generator()
    return model


@pytest.fixture(scope='session')
def netbatchnorm():
    model = SimpleBatchNormModel()
    return model


@pytest.fixture(scope='session')
def simpleconv():
    model = SimpleConv()
    return model


@pytest.fixture(scope='session')
def autoencoder():
    net = Autoencoder()
    return net


class RNNModel(nn.Module):
    def __init__(self, input_size, output_size, hidden_dim, n_layers):
        super().__init__()
        self.hidden_dim = hidden_dim
        self.n_layers = n_layers
        self.rnn = nn.RNN(input_size, hidden_dim, n_layers, batch_first=True)
        self.fc = nn.Linear(hidden_dim, output_size)

    def forward(self, x):
        batch_size = x.size(0)
        hidden = self.init_hidden(batch_size)
        out, hidden = self.rnn(x, hidden)
        out = out.contiguous().view(-1, self.hidden_dim)
        out = self.fc(out)
        return out, hidden

    def init_hidden(self, batch_size):
        hidden = torch.zeros(self.n_layers, batch_size, self.hidden_dim)
        return hidden


@pytest.fixture(scope='session')
def rnn():
    net = RNNModel(3, 1, 5, 3)
    return net


class MultiInputNet2(nn.Module):
    def __init__(self):
        super(MultiInputNet2, self).__init__()
        self.pool = nn.MaxPool2d(2, 2)
        self.conv1 = nn.Conv2d(3, 64, 3, padding=1)
        self.conv2 = nn.Conv2d(64, 96, 3, padding=1)
        self.conv3 = nn.Conv2d(96, 128, 3, padding=1)
        self.conv4 = nn.Conv2d(128, 192, 3, padding=1)
        self.avg_pool = nn.AdaptiveAvgPool2d((1, 1))
        self.linear1 = nn.Linear(192, 64)
        self.linear2 = nn.Linear(64, 4)

    def forward(self, image, unrefined_bounding_box):
        x = self.pool(F.relu(self.conv1(image)))
        x = self.pool(F.relu(self.conv2(x)))
        x = self.pool(F.relu(self.conv3(x)))
        x = F.relu(self.conv4(x))

        x = self.avg_pool(x)
        x = x.view(x.size(0), -1)
        x = F.relu(self.linear1(x))
        x = F.relu(self.linear2(x))
        x = x + unrefined_bounding_box
        return x


@pytest.fixture(scope='session')
def multi_input_net2():
    net = MultiInputNet2()
    return net


class LSTMModel(nn.Module):
    # https://www.deeplearningwizard.com/deep_learning/practical_pytorch/
    # pytorch_lstm_neuralnetwork/
    def __init__(self, input_dim, hidden_dim, layer_dim, output_dim):
        super(LSTMModel, self).__init__()
        self.hidden_dim = hidden_dim
        self.layer_dim = layer_dim
        self.lstm = nn.LSTM(input_dim, hidden_dim, layer_dim, batch_first=True)
        self.fc = nn.Linear(hidden_dim, output_dim)

    def forward(self, x):
        h0 = torch.zeros(
            self.layer_dim, x.size(0), self.hidden_dim
        ).requires_grad_()
        c0 = torch.zeros(
            self.layer_dim, x.size(0), self.hidden_dim
        ).requires_grad_()
        out, (hn, cn) = self.lstm(x, (h0.detach(), c0.detach()))
        out = self.fc(out[:, -1, :])
        return out


@pytest.fixture(scope='session')
def lstm_model():
    input_dim = 28
    hidden_dim = 100
    layer_dim = 1
    output_dim = 10
    model = LSTMModel(input_dim, hidden_dim, layer_dim, output_dim)
    return model



import sys
from dataclasses import dataclass
from functools import partial
from typing import List, IO, Union, Tuple, Type, Callable

import numpy as np
import torch
import torch.nn as nn

from torch.utils.hooks import RemovableHandle


__all__ = ('LayerInfo', 'NetworkInfo', 'inspect', 'summary')


InputShape = Tuple[int, ...]
OutputShape = List[Union[int, List[int]]]


@dataclass
class LayerInfo:
    name: str
    input_shape: List[int]
    output_shape: OutputShape
    trainable_params: int
    non_trainable_params: int


@dataclass
class NetworkInfo:
    total_params: int
    trainable_params: int
    total_input_size: int
    total_output_size: int
    total_params_size: int
    total_size: int


def make_network_info(
    info_list: List[LayerInfo],
    input_size: Union[InputShape, List[InputShape]],
    batch_size: int = 2,
) -> NetworkInfo:
    trainable_params = 0
    total_params = 0
    total_output = 0
    for layer in info_list:
        total_params += layer.trainable_params + layer.non_trainable_params
        total_output += abs(np.prod(layer.output_shape))
        trainable_params += layer.trainable_params

    # multiple inputs to the network
    if isinstance(input_size, tuple):
        input_size = [input_size]

    # assume 4 bytes/number (float on cuda).
    total_input_size = np.prod(sum(input_size, ())) * abs(batch_size) * 4
    # x2 for gradients
    total_output_size = 2 * total_output * 4
    total_params_size = total_params * 4
    total_size = total_params_size + total_output_size + total_input_size
    ni = NetworkInfo(
        total_params,
        trainable_params,
        total_input_size,
        total_output_size,
        total_params_size,
        total_size,
    )
    return ni


def should_attach_hook(model: nn.Module, module: nn.Module) -> bool:
    v = (
        not isinstance(module, nn.Sequential)
        and not isinstance(module, nn.ModuleList)
        and not (module is model)  # exclude self
    )
    return v


def _has_running_mean(module: nn.Module) -> bool:
    return (
        hasattr(module, 'running_mean')
        and hasattr(module.running_mean, 'size')
        and hasattr(module, 'track_running_stats')
        and module.track_running_stats
    )


def _has_running_var(module: nn.Module) -> bool:
    return (
        hasattr(module, 'running_var')
        and hasattr(module.running_var, 'size')
        and hasattr(module, 'track_running_stats')
        and module.track_running_stats
    )


class _ModuleHook:
    def __init__(self, batch_size: int):
        self.batch_size = batch_size
        self.layer_list: List[LayerInfo] = []

    def hook(
        self,
        module: nn.Module,
        input: torch.FloatTensor,
        output: torch.FloatTensor,
    ) -> None:

        # make layer name
        module_idx = len(self.layer_list)
        class_name = str(module.__class__).split('.')[-1].split("'")[0]
        name = f'{class_name}-{module_idx + 1}'

        # calculate input shape
        input_shape = list(input[0].size())
        output_shape = infer_shape(output)

        # calculate number of params
        trainable_params = 0
        non_trainable_params = 0

        for _, param in module.named_parameters():
            params = np.prod(param.size())
            if param.requires_grad:
                trainable_params += params
            else:
                non_trainable_params += params

        if _has_running_mean(module):
            params = np.prod(module.running_mean.size())
            non_trainable_params += params

        if _has_running_var(module):
            params = np.prod(module.running_var.size())
            non_trainable_params += params

        # recored result
        layer = LayerInfo(
            name,
            input_shape,
            output_shape,
            int(trainable_params),
            int(non_trainable_params),
        )
        self.layer_list.append(layer)


Output = Union[Tuple[torch.Tensor, ...], torch.Tensor]


def infer_shape(output: Output) -> OutputShape:
    shape: OutputShape = []
    if isinstance(output, (list, tuple)):
        shape = [infer_shape(o) for o in output]  # type: ignore
    else:
        shape = [int(s) for s in output.size()]
    return shape


def inspect(
    model: nn.Module,
    input_size: Union[InputShape, List[InputShape]],
    input_dtype: Type[torch.Tensor] = torch.FloatTensor,
    input_initializer: Callable[..., torch.Tensor] = torch.rand,
    batch_size: int = 2,
) -> List[LayerInfo]:
    hook = _ModuleHook(batch_size)
    handles: List[RemovableHandle] = []

    def register_hook(module: nn.Module) -> None:
        if should_attach_hook(model, module):
            h: RemovableHandle = module.register_forward_hook(
                hook.hook
            )
            handles.append(h)

    # multiple inputs to the network
    if isinstance(input_size, tuple):
        input_size = [input_size]

    # make fake input with batch_size of 2 for batchnorm
    x = [
        input_initializer(batch_size, *in_size).type(
            input_dtype  # type: ignore
        )
        for in_size in input_size
    ]
    # attach hooks to each applicable layer
    model.apply(register_hook)

    # forward pass
    try:
        model(*x)
    finally:
        # cleanup all attached hooks, to move model to original state
        for h in handles:
            h.remove()  # type: ignore

    return hook.layer_list


def summary(
    model: nn.Module,
    input_size: Union[InputShape, List[InputShape]],
    input_dtype: Type[torch.Tensor] = torch.FloatTensor,
    input_initializer: Callable[..., torch.Tensor] = torch.rand,
    batch_size: int = 2,
    file: IO[str] = sys.stdout,
    flush: bool = False,
) -> NetworkInfo:
    summary = inspect(
        model,
        input_size,
        input_dtype=input_dtype,
        input_initializer=input_initializer,
        batch_size=batch_size,
    )
    n = make_network_info(summary, input_size, batch_size)
    print_ = partial(print, file=file, flush=flush)
    print_('\n')
    print_('----------------------------------------------------------------')
    line_new = '{:>20}  {:>25} {:>15}'.format(
        'Layer (type)', 'Output Shape', 'Param #'
    )
    print_(line_new)
    print_('================================================================')

    for layer in summary:
        total_params = layer.trainable_params + layer.non_trainable_params
        line_new = '{:>20}  {:>25} {:>15}'.format(
            layer.name, str(layer.output_shape), '{0:,}'.format(total_params)
        )
        print_(line_new)
    MB = 1024.0 ** 2
    print_('================================================================')
    print_(f'Total params: {n.total_params:,}')
    print_(f'Trainable params: {n.trainable_params:,}')
    print_(f'Non-trainable params: {n.total_params - n.trainable_params:,}')
    print_('----------------------------------------------------------------')
    print_(f'Input size (MB): {n.total_input_size / MB:0.2f}')
    print_(f'Forward/backward pass size (MB): {n.total_output_size / MB:0.2f}')
    print_(f'Params size (MB): {n.total_params_size / MB :0.2f}')
    print_(f'Estimated Total Size (MB): {n.total_size / MB:0.2f}')
    print_('----------------------------------------------------------------')
    return n



import os
import re
import sys
from setuptools import setup, find_packages


install_requires = ['torch>=1.1.0']

PY36 = (3, 6, 0)
PY37 = (3, 7, 0)


if sys.version_info < PY36:
    raise RuntimeError('torch-inspect requires Python 3.6.0+')


if sys.version_info < PY37:
    install_requires.append('dataclasses==0.6')


def read(f):
    return open(os.path.join(os.path.dirname(__file__), f)).read().strip()


def read_version():
    regexp = re.compile(r"^__version__\W*=\W*'([\d.abrc]+)'")
    init_py = os.path.join(
        os.path.dirname(__file__), 'torch_inspect', '__init__.py'
    )
    with open(init_py) as f:
        for line in f:
            match = regexp.match(line)
            if match is not None:
                return match.group(1)
        else:
            raise RuntimeError(
                'Cannot find version in torch_inspect/__init__.py'
            )


classifiers = [
    'License :: OSI Approved :: Apache Software License',
    'Intended Audience :: Developers',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Operating System :: OS Independent',
    'Development Status :: 3 - Alpha',
]


setup(
    name='torch-inspect',
    version=read_version(),
    description=('Utility functions that prints a summary of a model.'),
    long_description='\n\n'.join((read('README.rst'), read('CHANGES.rst'))),
    classifiers=classifiers,
    platforms=['POSIX'],
    author='Nikolay Novik',
    author_email='nickolainovik@gmail.com',
    url='https://github.com/jettify/pytorch-inspect',
    download_url='https://pypi.org/project/torch-inspect/',
    license='Apache 2',
    packages=find_packages(),
    install_requires=install_requires,
    keywords=['torch-inspect', 'pytorch'],
    zip_safe=True,
    include_package_data=True,
)
